/*
 * File: linux/security/gtsec.c
 * Author: Gerrit Htzel (gt@syndicus.org)
 * Date: Dec 24, 2003
 * (last updated: 16.05.2004)
 *
 * Description:
 *   LSM module.
 *   Currently these security modes are supported.
 *   They can be activated separately, but only one of each category at the same time
 *   for the same task.
 *     Network Security:
 *     * vnetonly
 *      Network access is restricted to IPs IN vnet_list and NOT IN novnet_list.
 *      Has immediate effect.
 *          Options (call BEFORE calling vnetonly)
 *            * vnet_ip a.b.c.d OR a.b.c.d/A.B.C.D OR a.b.c.d/E
 *              IPs that are in vnet (IP/NETMASK).
 *              valid examples:
 *                e.g.: echo -n "vnet_ip 192.168.1.1" > /proc/$$/attr/exec
 *                e.g.: echo -n "vnet_ip 192.168.1.0/24" > /proc/$$/attr/exec
 *                e.g.: echo -n "vnet_ip 192.168.1.0/255.255.255.0" > /proc/$$/attr/exec
 *            * novnet_ip a.b.c.d OR a.b.c.d/A.B.C.D OR a.b.c.d/E
 *              IPs that are not in vnet, even when listed in vnet_ip (IP/NETMASK).
 *                e.g.: echo -n "novnet_ip 192.168.1.27/255.255.255.255" > /proc/$$/attr/exec
 *     * allbutvnet
 *          Is basically the same as vnetonly, but inverts the policy: A connections is
 *          allowed if it would be denied by vnetonly.
 *     Local Security:
 *       * bsdjail
 *          What Serge Hallyn implemented with his bsdjail LSM.
 *          Unlike Hallyn's bsdjail version though, gtsec will go directly into
 *          bsdjail mode after executing 'echo -n "bsdjail" > ...'.
 *          However, you still should call "exec /bin/sh" or something after that:
 *            echo -n "bsdjail /chroot" > /proc/$$/attr/exec
 *            exec /bin/sh
 *        * softjail
 *          Access to /home and /root is denied completly.
 *          Write access is limited to:
 *            /tmp - Many programs will need this.
 *            /var - As do other need this, e.g. exim when sending mail
 *                   (log and spool files).
 *            /dev - e.g. /dev/dsp /dev/mixer
 *          Only tasks running within the same jail are visible.
 *          Has immediate effect.
 *          Note: Currently writing to /proc is also denied. So you cannot activate
 *          a lan_sectype mode *after* entering softjail mode. This might be changed soon.
 *          You might want to do a
 *            export HOME=/tmp/newhome
 *          Options (call BEFORE calling softjail)
 *            * sjhome /HOME/DIR
 *              Mandatory. Set home directory within jail.
 *            * sjuser USERNAME
 *              Mandatory. Set username within jail.
 *            * ip x.x.x.x
 *              Bind to this IP is allowed, everything else denied.
 *     Log Mode:
 *       * netlog
 *         Loggs all network connections made
 *       Has immediate effect.
 *       Very experimental.
 *
 *
 * History:
 *   Based upon the bsdjail LSM by Serge Hallyn (serue@us.ibm.com)
 *   Tried to apply QualityCode: Do the simple thing, even when it hurts.
 *
 * To do:
 *  * Softjails can see each other's home directories in /tmp/... !
 *  * Smarter handling of file/dir renaming !
 *  * What about broadcast packets ?
 *  * In vnetonly mode deny all incoming connections from not-vnet and
 *    accordingly for allbutvnet !?
 *  * Limit rw perm to /dev/
 *  * What about mount --bind ? It seems to be properly handled in softjail
 *     mode. But haven't really thought about it, yet.
 *  * When gtsec is not loeaded symlink /etc/passwd points to nowhere.
 *  * Handle security type (ro/rw/none) of directories dynamically.
 *  * Include a locking mechanism that denies writing to /proc/gtsec/passwd.softjail
 *  * Only show network connections from current device in bsdjail (netstat -an)
 *  * 'touch /home' works in softjail mode as root. That okay?
 *
 *
 * How to use:
 *     0. Once, if you want to use softjail mode:
 *        mv /etc/passwd /etc/passwd.original && ln -s /etc/passwd /proc/gtsec/passwd
 *     1. modprobe gtsec [gtsec_debug=1]
 *     2. Once, after loading:
 *        cat /etc/passwd > /proc/gtsec/passwd.softjail
 *     2. Example: Allow connections to your LAN only, no internet:
 *          echo -n "vnet_ip 127.0.0.1/255.0.0.0"      > /proc/$$/attr/exec
 *          echo -n "vnet_ip 0.0.0.0"                  > /proc/$$/attr/exec
 *          echo -n "vnet_ip 192.168.1.0/24"           > /proc/$$/attr/exec
 *          echo -n "novnet_ip 192.168.1.110"          > /proc/$$/attr/exec
 *          echo -n "novnet_ip 192.168.1.111"          > /proc/$$/attr/exec
 *          echo -n "novnet_ip 192.168.1.112"          > /proc/$$/attr/exec
 *          echo -n "vnetonly"                         > /proc/$$/attr/exec
 *     3. Example: You may want (additionally) to have no access to the
 *                 home directories and read-only for the rest of the fs
 *                 (besides some exceptions, see above):
 *         echo -n "sjhome /tmp/my_new_home"        > /proc/$$/attr/exec
 *         echo -n "sjuser my_new_username"         > /proc/$$/attr/exec
 *         echo -n "softjail"                       > /proc/$$/attr/exec
 *
 * Copyright (C) 2004 Gerrit Htzel (gt@syndicus.org)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 */

// TODO: more includes than we need, I suppose
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/security.h>
#include <linux/mount.h>
#include <linux/namei.h>
#include <linux/namespace.h>
#include <linux/proc_fs.h>
#include <linux/in.h>
#include <linux/pagemap.h>
#include <linux/ip.h>
#include <asm/uaccess.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/seq_file.h>
#include <linux/un.h>
#include <net/af_unix.h>

#include <linux/stat.h>


#ifndef CONFIG_SECURITY
	#error You have to compile your kernel with the CONFIG_SECURITY option set.
#endif

static int gtsec_debug = 0;

/* kmalloc is limited to 128 kB.
 * large kmallocs (I'd guess: > 16 kB) can fail due to memory fragmentation after long uptimes!
 */
//static int max_passwd_len = 16384;		/* This INCLUDES the NULL-terminator! */
static int max_passwd_len = 4096;		/* This INCLUDES the NULL-terminator! */

MODULE_PARM(gtsec_debug, "i");
MODULE_PARM_DESC(gtsec_debug, " Print GTSEC debugging messages.");
MODULE_PARM(max_passwd_len, "i");
MODULE_PARM_DESC(max_passwd_len, " Maximum size of passwd file.");

#define DBG 0
#define WARN 1
#define print_debug(how, fmt, arg... ) \
	do { \
		if ( how || gtsec_debug ) \
			printk(KERN_NOTICE "%s: %s: " fmt, \
				MY_NAME, __FUNCTION__, \
				## arg ); \
	} while ( 0 )

#define print_netlog(fmt, arg... ) \
		if (tsec->logtype == NETLOG) \
			printk(KERN_DEBUG "netlog: " fmt, ## arg );

/* convert IP a.b.c.d to __u32 */
#define get_ip4(a,b,c,d) (a | b << 8 | c << 16 | d << 24)

#define checkip			list[i]
#define checkmask		list[i+1]

/* flag to keep track of how we were registered */
static int secondary = 0;

#define NONE					0
/* modes for tsec.net_sectype: */
#define VNETONLY				1
#define ALLBUTVNET				2
/* modes for tsec.loc_sectype: */
#define BSDJAIL					3
#define SOFTJAIL				4
/* modes for logtype: */
#define NETLOG					5

/* options */
#define OPT_SJUSER				6		/* softjail: set the userNAME used in jail */
#define OPT_SJHOME				7		/* softjail: set the user's home directory used in jail */

#define OPT_VNET_IP				8		/* softjail */
#define OPT_NOVNET_IP			9		/* softjail */

#define OPT_ROOT			   10		/* bsdjail */
#define OPT_IP				   11		/* bsdjail */

#define OPT_TIMESLICE		   12
#define OPT_SLICE			   13
#define OPT_NRTASK			   14
#define OPT_MEMLOCK			   15
#define OPT_DATA			   16
#define OPT_NICE			   17


/*
 * The task structure holding gtsec information.
 * Taskp->security points to one of these (or is null).
 * There is exactly one gtsec_task_sec for each jail.  If >1 process
 * are in the same jail, they share the same gtsec_task_sec.
 */
struct gtsec_task_sec {
	short in_use;					/* in_use:
									 * if 0, then this task is actually setting up a jail,
									 * not currently in one. if 1, this task is in a jail of some
									 * sectype.
									 */
	atomic_t refcount;				/* how many processes in this jail */

	short net_sectype;				/* network type of security */
	short loc_sectype;				/* local type of security */
	short logtype;					/* type of logging */

	/* bsdjail: these are set on writes to /proc/<pid>/attr/exec */
	char *root_pathname;			/* char * containing path to use as jail root */
	__u32 bsdj_ip;					/* ip addr to use for jail (a.k.a. realaddr) */

	/* bsdjail: these are set when a jail becomes active */
	struct dentry *dentry;			/* dentry of fs root */
	struct vfsmount *mnt;			/* vfsmnt of fs root */

	char *sj_home;					/* char * containing path to softjail homedirectory */
	char *sj_username;				/* char * containing the name of the user within the jail */
	uid_t sj_uid;					/* the uid of the user within the jail */
	gid_t sj_gid;					/* the gid of the user within the jail */


	/*
	 * For the sake of simplicity we make static arrays.
	 * Besides, the code to do dynamic allocation alone will get bigger than 10 arrays.
	 * Besides, dynamic allocation is always error prone.
	 */
	#define MAX_IPLIST	15
	__u32 vnet_list[MAX_IPLIST];	/* list of IPs that are in the virutal net*/
	__u32 novnet_list[MAX_IPLIST];	/* list of IPs taht are not in vnet, although in vnet_list */

	/* Resource limits.  0 = no limit */
	long max_nrtask;	/* maximum number of tasks within this jail. */
	long cur_nrtask;	/* current number of tasks within this jail. */
	long maxtimeslice;  /* max timeslice in ms for procs in this jail */
	long nice;      	/* nice level for processes in this jail */
	long max_data, max_memlock;  /* equivalent to RLIMIT_{DATA,MEMLOCK} */
};

unsigned long i_home = 0;	// TODO: initialize to what to indicate invalid/unchecked inode?
unsigned long i_root = 0;	//
unsigned long i_tmp = 0;	//
unsigned long i_var = 0;	//
unsigned long i_dev = 0;	//

/* For password and proc handling: */
static struct proc_dir_entry *home_symlink, *passwd_symlink, *passwd_file, *gtsec_proc_dir;
static char *passwd_data = NULL;
static rwlock_t place_lock = RW_LOCK_UNLOCKED;
#define ORIGINAL_ETC_PASSWD "/etc/passwd.original"


/* allow use with stacker LSM (not really tested) */
#define get_security(st,p,type) (p->type)
#define set_security(st,p,type,data) (p->type = data)

#define jail_of(proc) (get_security(task,proc,security))

#define MY_NAME "gtsec"



static inline int
in_jail(struct task_struct *t)
{
	struct gtsec_task_sec *tsec = get_security(task,t,security);
	
	if (tsec && tsec->in_use)
		return 1;

	return 0;
}

/*
 * alloc_task_security and free_task_security:
 */
static struct gtsec_task_sec *
alloc_task_security(struct task_struct *tsk)
{
	struct gtsec_task_sec *tsec;

	tsec = kmalloc(sizeof(struct gtsec_task_sec), GFP_KERNEL);
	if (!tsec)
		return ERR_PTR(-ENOMEM);
	memset(tsec, 0, sizeof(struct gtsec_task_sec));
	set_security(task,tsk,security,tsec);
	return tsec;
}

static void
free_task_security(struct task_struct *tsk)
{
	struct gtsec_task_sec *tsec;

	tsec = get_security(task,tsk,security);
	if (!tsec) return;
	if (tsec->root_pathname) kfree(tsec->root_pathname);
	if (tsec->sj_home) kfree(tsec->sj_home);
	//if (tsec->vnet_list) kfree(tsec->vnet_list);
	//if (tsec->novnet_list) kfree(tsec->novnet_list);
	kfree(tsec);
	set_security(task,tsk,security,NULL);
}



/*
 * Called when a process is placed into a new jail to handle the
 * actual creation of the jail.
 */
int create_jail(struct task_struct *tsk)
{
	struct nameidata nd;
	struct gtsec_task_sec *tsec;
	int retval = -EFAULT;

	tsec = get_security(task,tsk,security);
	if (!tsec || !(tsec->net_sectype || tsec->loc_sectype || tsec->logtype))
		goto out;

	if (tsec->loc_sectype == BSDJAIL) {
		/* 
		 * USE_JAIL_NAMESPACE: could be useful, so that future mounts outside
		 * the jail don't affect the jail.  But it's not necessary, and
		 * requires exporting copy_namespace from fs/namespace.c
#define USE_JAIL_NAMESPACE
		 */
#ifdef USE_JAIL_NAMESPACE
		print_debug(DBG, "gtsec: copying namespace.\n");
		retval = -EPERM;
		if (copy_namespace(CLONE_NEWNS, tsk))
			goto out;
		print_debug(DBG, "gtsec: copied namespace.\n");
#endif

		/* find our new root directory */
		retval = path_lookup(tsec->root_pathname, LOOKUP_FOLLOW | LOOKUP_DIRECTORY, &nd);
		if (retval) goto out;

		/* and set the fsroot to it */
		set_fs_root(tsk->fs, nd.mnt, nd.dentry);
		set_fs_pwd(tsk->fs, nd.mnt, nd.dentry);
		print_debug(DBG, "gtsec: New root has been set to: \"%s\".\n", tsec->root_pathname);

		tsec->mnt = mntget(nd.mnt);
		tsec->dentry = dget(nd.dentry);
		path_release(&nd);
	}
	else if (tsec->loc_sectype == SOFTJAIL) {

		/* Initially get the inodes of the directories we need to watch */

		if (!i_home) {
			retval = path_lookup("/home", LOOKUP_FOLLOW | LOOKUP_DIRECTORY, &nd);
			if (retval) goto out;
			i_home=nd.dentry->d_inode->i_ino;
		}
		if (!i_root) {
			retval = path_lookup("/root", LOOKUP_FOLLOW | LOOKUP_DIRECTORY, &nd);
			if (retval) goto out;
			i_root=nd.dentry->d_inode->i_ino;
		}
		if (!i_tmp) {
			retval = path_lookup("/tmp", LOOKUP_FOLLOW | LOOKUP_DIRECTORY, &nd);
			if (retval) goto out;
			i_tmp=nd.dentry->d_inode->i_ino;
		}
		if (!i_var) {
			retval = path_lookup("/var", LOOKUP_FOLLOW | LOOKUP_DIRECTORY, &nd);
			if (retval) goto out;
			i_var=nd.dentry->d_inode->i_ino;
		}
		if (!i_dev) {
			retval = path_lookup("/dev", LOOKUP_FOLLOW | LOOKUP_DIRECTORY, &nd);
			if (retval) goto out;
			i_dev=nd.dentry->d_inode->i_ino;
		}

	}

	if (!in_jail(current)) {
		/* No jail set up for this task, yet. */
		atomic_inc(&tsec->refcount);
		tsec->in_use = 1;

		/* won't let ourselves be removed until this jail goes away */
		try_module_get(THIS_MODULE);
	}

	print_debug(DBG, "gtsec: jail created/modified for %s.\n", current->comm);

	return 0;

out:
	return retval;
}

static void
disable_jail(struct gtsec_task_sec *tsec)
{
	/*
	 *   don't need to put namespace, it will be done automatically
	 *     when the last process in jail is put.
	 *   DO need to put the dentry and vfsmount
	 */

	dput(tsec->dentry);
	mntput(tsec->mnt);
	module_put(THIS_MODULE);
	print_debug(DBG, "jail disabled for %s.\n", current->comm);
}



///////////////////////////////////////////////////////////////////////////////
// LSM /proc/<pid>/attr hooks.
///////////////////////////////////////////////////////////////////////////////

int
add_tsec_ip(__u32 *list, const __u32 ip, const __u32 mask)
{

	int i;
	for (i=0; i <= MAX_IPLIST; i+=2)
	{
		if (checkip || checkmask) {
			continue;
		}

		checkip = ip;
		checkmask = mask;
		return 1;
	}

	/* More than MAX_IPLIST !!! */
	return -EINVAL;
}

/* allocate memory and store value */
static
//inline
int
get_mykopt(char **var, const void *value, const int size, const size_t len)
{

	if (*var) kfree(*var);
	*var = kmalloc(size-len, GFP_KERNEL);
	if (!*var) return -ENOMEM;
	strncpy(*var, value+len+1, size-len-1);
	(*var)[size-len-1] = '\0';

	return 1;
}

/* return value as long */
static
//inline
long
get_mylopt(const void *value, const int size, const size_t len)
{
	long val;
	char *opt = NULL;	/* NULL, or get_mykopt wants to kfree ! */
	val = get_mykopt(&opt, value, size, len);
	if (val > 0) {
		val = simple_strtoul(opt, NULL, 0);
		kfree (opt);
	}
	return val;
}

/*
 * Convert value to ip/netmask.
 * Valid are examples:
 *   192.168.1.1
 *   192.168.1.0/24
 *   192.168.1.0/255.255.255.0
 */
static int
get_myipopt(const void *value, const int size, const size_t len, __u32 *ip, __u32 *mask)
{
		char *ip_name = NULL;	/* NULL, or get_mykopt wants to kfree ! */
		unsigned int a,b,c,d;
		unsigned int A,B,C,D;
		int ret;

		ret = get_mykopt(&ip_name, value, size, len);
		if (ret<0) return ret;

		if (sscanf(ip_name,"%u.%u.%u.%u/%u.%u.%u.%u",&a,&b,&c,&d,&A,&B,&C,&D) == 8 &&
			(a<=255 && b<=255 && c<=255 && d<=255 && A<=255 && B<=255 && C<=255 && D<=255) ) {
			*ip = get_ip4(a,b,c,d);
			*mask = get_ip4(A,B,C,D);
		}
		else if (sscanf(ip_name,"%u.%u.%u.%u/%u",&a,&b,&c,&d,&A) == 5 &&
				(a<=255 && b<=255 && c<=255 && d<=255) ) {
			if (A > 32) {
				kfree (ip_name);
				return -EINVAL;
			}
			A = htonl ( 0xFFFFFFFFUL << (32 - A) & 0xFFFFFFFFUL );
			*ip = get_ip4(a,b,c,d);
			*mask = A;
		}
		else if (sscanf(ip_name,"%u.%u.%u.%u",&a,&b,&c,&d) == 4 &&
				(a<=255 && b<=255 && c<=255 && d<=255) ) {
			*ip = get_ip4(a,b,c,d);
			*mask = get_ip4(255,255,255,255);
		}
		else {
			kfree (ip_name);
			return -EINVAL;
		}

		kfree (ip_name);
		return size;
}

/*
 * You may write into /proc/<pid>/attr/exec.
 */
static int
gtsec_setprocattr(struct task_struct *p, char *name, void *value, size_t size)
{
	struct gtsec_task_sec *tsec;
	int ret;
	short requested=NONE;
	long val;
	__u32 theip;
	__u32 themask;

	/*
	 * Current process may write to /proc/<pid>/attr/exec.
	 */
	if (p != current || strcmp(name, "exec"))
		return -EPERM;

	if (strncmp(value, "vnetonly", 8)==0) requested = VNETONLY;
	else if (strncmp(value, "allbutvnet", 10)==0) requested = ALLBUTVNET;
	else if (strncmp(value, "softjail", 8)==0) requested = SOFTJAIL;
	else if (strncmp(value, "bsdjail ", 7)==0) requested = BSDJAIL;
	else if (strncmp(value, "root ", 4)==0) requested = OPT_ROOT;
	else if (strncmp(value, "ip ", 2)==0) requested = OPT_IP;
	else if (strncmp(value, "netlog", 6)==0) requested = NETLOG;
	else if (strncmp(value, "sjuser", 6)==0) requested = OPT_SJUSER;
	else if (strncmp(value, "sjhome", 6)==0) requested = OPT_SJHOME;
	else if (strncmp(value, "vnet_ip", 7)==0) requested = OPT_VNET_IP;
	else if (strncmp(value, "novnet_ip", 9)==0) requested = OPT_NOVNET_IP;
	else if (strncmp(value, "slice", 5)==0) requested = OPT_SLICE;
	else if (strncmp(value, "timeslice", 9)==0) requested = OPT_TIMESLICE;
	else if (strncmp(value, "nrtask", 6)==0) requested = OPT_NRTASK;
	else if (strncmp(value, "memlock", 7)==0) requested = OPT_MEMLOCK;
	else if (strncmp(value, "data", 4)==0) requested = OPT_DATA;
	else if (strncmp(value, "nice", 4)==0) requested = OPT_NICE;

	tsec = get_security(task,current,security);

	if (tsec && tsec->in_use) {
		/* If lan_sectype not set, task can set lan_sectype but not loc_sectype.
		 * If loc_sectype not set, task can set loc_sectype but not net_sectype.
		 */
		if (tsec->net_sectype &&
			(requested == VNETONLY || requested == ALLBUTVNET))
			return -EINVAL;  /* let them guess why */

		if (tsec->loc_sectype &&
			(requested == BSDJAIL || requested == SOFTJAIL))
			return -EINVAL;  /* let them guess why */

		/* Ignore NETLOG which may be switched WITHING jail. */
	}

	if (!tsec)
		tsec = alloc_task_security(current);
	if (IS_ERR(tsec))
		return -ENOMEM;
		

	if (requested == OPT_SJUSER) {
		tsec->sj_uid = current->euid;
		tsec->sj_gid = current->egid;

		/* For security reasons: Deny username "root*" for non-root. */
		if ( (tsec->sj_uid != 0 || tsec->sj_gid != 0)
			&& strncmp(value+6+1, "root", 4) == 0) return -EPERM;

		ret = get_mykopt(&(tsec->sj_username), value, size, 6);
		if (ret<0) return ret;

		print_debug(DBG, "option: username: \"%s\" (%d:%d).\n",
				tsec->sj_username, tsec->sj_uid, tsec->sj_gid);

		return size;
	}
	else if (requested == OPT_SJHOME) {
		ret = get_mykopt(&(tsec->sj_home), value, size, 6);
		if (ret<0) return ret;

		print_debug(DBG, "option: home directory: \"%s\".\n", tsec->sj_home);

		return size;
	}
	else if (requested == OPT_VNET_IP) {
		if ( (ret = get_myipopt(value, size, 7, &theip, &themask)) > 0 ) {
			add_tsec_ip(tsec->vnet_list, theip, themask);
			print_debug(DBG, "vnet_ip: %u.%u.%u.%u/%u.%u.%u.%u\n", NIPQUAD(theip), NIPQUAD(themask));
		}
		return ret;
	}
	else if (requested == OPT_NOVNET_IP) {
		if ( (ret = get_myipopt(value, size, 9, &theip, &themask)) > 0 ) {
			add_tsec_ip(tsec->novnet_list, theip, themask);
			print_debug(DBG, "novnet_ip: %u.%u.%u.%u/%u.%u.%u.%u\n", NIPQUAD(theip), NIPQUAD(themask));
		}
		return ret;
	}
	else if (requested == OPT_ROOT) {
		ret = get_mykopt(&(tsec->root_pathname), value, size, 4);
		if (ret<0) return ret;
		print_debug(DBG, "option: bsdjail root: \"%s\".\n", tsec->root_pathname);
		return size;
	}
	else if (requested == OPT_IP) {
		if ( (ret = get_myipopt(value, size, 2, &theip, &themask)) > 0 ) {
			tsec->bsdj_ip = theip;
			print_debug(DBG, "bsdjail ip: %u.%u.%u.%u\n", NIPQUAD(theip));
		}
		return ret;
	}
	else if (requested == OPT_TIMESLICE) {
		if ( (val = get_mylopt(value, size, 9)) ) {
			tsec->maxtimeslice = val;
			print_debug(DBG, "option: timeslice: %ld\n", val);
		}
		return size;
	}
	else if (requested == OPT_SLICE) {
		if ( (val = get_mylopt(value, size, 5)) ) {
			tsec->maxtimeslice = val;
			print_debug(DBG, "option: timeslice: %ld\n", val);
		}
		return size;
	}
	else if (requested == OPT_NRTASK) {
		if ( (val = get_mylopt(value, size, 6)) ) {
			tsec->maxtimeslice = val;
			print_debug(DBG, "option: max_nrtask: %ld\n", val);
		}
		return size;
	}
	else if (requested == OPT_MEMLOCK) {
		if ( (val = get_mylopt(value, size, 7)) ) {
			tsec->maxtimeslice = val;
			print_debug(DBG, "option: max_memlock: %ld\n", val);
		}
		return size;
	}
	else if (requested == OPT_DATA) {
		if ( (val = get_mylopt(value, size, 4)) ) {
			tsec->maxtimeslice = val;
			print_debug(DBG, "option: max_data: %ld\n", val);
		}
		return size;
	}
	else if (requested == OPT_NICE) {
		if ( (val = get_mylopt(value, size, 4)) ) {
			tsec->maxtimeslice = val;
			print_debug(DBG, "option: nice: %ld\n", val);
		}
		return size;
	}
	else if (requested == VNETONLY) {
		/* vnet_list and novnet_list are already initialised to 0.*/
		tsec->net_sectype = VNETONLY;
	}
	else if (requested == ALLBUTVNET)
		tsec->net_sectype = ALLBUTVNET;
	else if (requested == NETLOG)
		tsec->logtype = NETLOG;
	else if (requested == SOFTJAIL) {
		if (!tsec->sj_home || tsec->sj_home[0] == '\0') return -EINVAL;
		if (!tsec->sj_username || tsec->sj_username[0] == '\0') return -EINVAL;
		tsec->loc_sectype = SOFTJAIL;
	}
	else if (requested == BSDJAIL) {
		if (!tsec->root_pathname || tsec->root_pathname[0] == '\0') return -EINVAL;
		tsec->loc_sectype = BSDJAIL;
	}
	else if (requested == NETLOG) {
		/* Switch */
		//tsec->logtype = (tsec->logtype==NETLOG)?0:1;
		tsec->logtype = 1;
	}
	else return -EINVAL;

	ret = create_jail(current);
	if (ret) {
		/* We failed. */
		free_task_security(current);
		return -EINVAL;
	}

	return size;
}


/*
 * LSM /proc/<pid>/attr read hook.
 */
static int
gtsec_getprocattr(struct task_struct *p, char *name, void *value, size_t size)
{
	struct gtsec_task_sec *tsec;
	int err = 0;

	if (in_jail(current))
		return -EINVAL;  /* let them guess why */
		
	/* cat /proc/$$/attr/exec: */
	if (strcmp(name, "exec") == 0) {
		/* Print usage some help */
		err = snprintf(value, size,
			"Valid keywords:\n"
			"general (valid in all security types; active on next exec):\n"
			"  nrtask      <max number of tasks in this jail>\n"
			"  nice        <nice level for processes in this jail>\n"
			"  [time]slice <max timeslice per process in msecs>\n"
			"  data        <max data size per process in bytes>\n"
			"  memlock     <max lockable memory per process in bytes>\n"
			"\n"
			"vnetonly (allow network traffic to virtual net):\n"
			"  vnet_ip    <this ip/mask is within virtual network>\n"
			"  novnet_ip  <this ip/mask is not within virtual network>\n"
			"allbutvnet (deny network traffic to virtual net):\n"
			"  same options as for vnetonly\n"
			"\n"
			"softjail:\n"
			"  sjuser      <username in softjail>\n"
			"  sjhome      <home directory in softjail>\n"
			"  ip          <bind ip4-addr>\n"
			"bsdjail:\n"
			"  root        <pathname>\n"
			"  ip          <bind ip4-addr>\n"
			"\n"
			"netlog (experimental, do not use):\n"
			"  no options\n");
		return err;
	}

	if (strcmp(name, "current") != 0)
		return -EPERM;

	tsec = get_security(task, p, security);
	if (!tsec || !tsec->in_use) {
		err = snprintf(value, size, "Not Jailed\n");
	} else {
		//err = snprintf(value, size, "Jailed.\n");
		err = snprintf(value, size,
			"TODO...\n"
			"Root: %s\n"
			//"IP: %u.%u.%u.%u\n"
			"max_nrtask %lu current nrtask %lu max_timeslice %lu nice %lu\n"
			"max_memlock %lu max_data %lu\n",
			tsec->root_pathname,
			//(tsec->bsdj_ip ? (NIPQUAD(tsec->bsdj_ip)) : (0,0,0,0)),
			tsec->max_nrtask, tsec->cur_nrtask, tsec->maxtimeslice,
			tsec->nice, tsec->max_data, tsec->max_memlock);
	}

	return err;
}

///////////////////////////////////////////////////////////////////////////////
// END: LSM /proc/<pid>/attr hooks.
///////////////////////////////////////////////////////////////////////////////




/*
 * Forbid a process in a jail from sending a signal to a process in another
 * (or no) jail through file sigio.
 *
 * We consider the process which set the fowner to be the one sending the
 * signal, rather than the one writing to the file.  Therefore we store the
 * jail of a process during jail_file_set_fowner, then check that against
 * the jail of the process receiving the signal.
 */
static int
jail_file_send_sigiotask(struct task_struct *tsk, struct fown_struct *fown,
       int fd, int reason)
{
	struct file *file;
	struct bsdjail_task_sec *tsec;

	// TODO : check for SIGIO here!

	if (!in_jail(current))
		return 0;

	file = (struct file *)((long)fown - offsetof(struct file,f_owner));
	tsec = jail_of(tsk);

/*	if (jail_of(tsk) != jail_of(current))*/
	if (get_security(file,file,f_security) != tsec) {
		print_debug(DBG, "Send sigio denied.\n");
		return -EPERM;
	}

	return 0;
}

static int
jail_file_set_fowner(struct file *file)
{
	struct bsdjail_task_sec *tsec;

	tsec = jail_of(current);
	set_security(file,file,f_security,tsec);

	return 0;
}

/* 
 * LSM ptrace hook:
 * process in jail may not ptrace process not in the same jail
 */
static int
gtsec_ptrace (struct task_struct *doctor, struct task_struct *patient)
{
	if (in_jail(doctor)) {
		if (jail_of(doctor) == jail_of(patient))
			return 0;
		return -EPERM;
	}
	return 0;
}


#ifdef CONFIG_SECURITY_NETWORK


static inline int
is_listed(const __u32 sin_addr, const __u32 *list)
{
	int i;
	for (i=0; (checkip || checkmask) && (i <= MAX_IPLIST); i+=2)
	{
		if ( (sin_addr & checkmask) == (checkip & checkmask) )
			return 1;
	}
	return 0;
}
#define is_vnetip(sin_addr)		is_listed(sin_addr,tsec->vnet_list)
#define is_novnetip(sin_addr)	is_listed(sin_addr,tsec->novnet_list)
#define is_vnet(sin_addr)		( is_vnetip(sin_addr) && !is_novnetip(sin_addr) )

static int
gtsec_socket_connect(struct socket *sock, struct sockaddr *address, int addrlen)
{
	struct sockaddr_in *addr;
	__u32 sin_addr;
	struct gtsec_task_sec *tsec;
	//unsigned short int sinport;
	unsigned short sport;

	if (!in_jail(current)) return 0;

	/*
	 * No support for IPv6, yet. So deny.
	 * This list should be extended by e.g. IPX.
	 */
	if (sock->sk->sk_family == AF_INET6)
		return -EPERM;

	if (sock->sk->sk_family != AF_INET)
		return 0;

	addr = (struct sockaddr_in *)address;
	sin_addr = addr->sin_addr.s_addr;
	//printk (KERN_INFO "GTSEC: %s wants to connect to: %u.%u.%u.%u\n", current->comm, NIPQUAD(sin_addr));

	sport = ntohs(addr->sin_port);

	tsec = get_security(task,current,security);
	if (!tsec->net_sectype) goto grant;

	if ( is_vnet(sin_addr) ) {
		if (tsec->net_sectype == VNETONLY) goto grant;
		goto deny;
	}
	else if (tsec->net_sectype == ALLBUTVNET) goto grant;

deny:
	print_netlog ("Denied %s to connect to: %u.%u.%u.%u:%d\n", current->comm, NIPQUAD(sin_addr), sport);
	return -EPERM;

grant:
	print_netlog ("Granted %s to connect to: %u.%u.%u.%u:%d\n", current->comm, NIPQUAD(sin_addr), sport);
	return 0;
}

static int
gtsec_socket_sendmsg(struct socket * sock, struct msghdr * msg, int size)
{
	struct sockaddr_in *addr;
	__u32 sin_addr;
	struct gtsec_task_sec *tsec;
	unsigned short sport;

	if (!in_jail(current)) return 0;

	/* The socket is connected. Security already checked by socket_connect. */
	if ( !(msg->msg_name) ) return 0;

	addr = (struct sockaddr_in*)(msg->msg_name);

	sin_addr = addr->sin_addr.s_addr;
	sport = ntohs(addr->sin_port);

	/*
	 * No support for IPv6, yet. So deny.
	 * TODO: This list should be extended by e.g. AF_IPX. Or deny all by default?
	 */
	if (sock->sk->sk_family == AF_INET6)
		return -EINVAL;

	if (sock->sk->sk_family != AF_INET)
		return 0;

	//printk (KERN_INFO "GTSEC: %s wants to connect to: %u.%u.%u.%u\n", current->comm, NIPQUAD(sin_addr));

	tsec = get_security(task,current,security);
	if (!tsec->net_sectype) goto grant;

	else if ( is_vnet(sin_addr) ) {
		if (tsec->net_sectype == VNETONLY) goto grant;
		goto deny;
	}
	else if (tsec->net_sectype == ALLBUTVNET) goto grant;

deny:
	print_netlog ("Denied %s to sendmsg to: %u.%u.%u.%u:%d\n", current->comm, NIPQUAD(sin_addr), sport);
	/* TODO: Which error to return ? */
	return -EINVAL;

grant:
	print_netlog ("Granted %s to sendmsg to: %u.%u.%u.%u:%d\n", current->comm, NIPQUAD(sin_addr), sport);
	return 0;

}


#define loopbackaddr htonl((127 << 24) | 1)

/*
 * process in jail may only use one (aliased) ip address.  If they try to
 * attach to 127.0.0.1, that is remapped to their own address.  If some
 * other address (and not their own), deny permisison
 */
static int gtsec_socket_unix_bind(struct socket *sock, struct sockaddr *address,
		int addrlen);

static int
gtsec_socket_bind(struct socket *sock, struct sockaddr *address, int addrlen)
{
	struct gtsec_task_sec *tsec;
	struct sockaddr_in *inaddr;
	__u32 sin_addr, jailaddr;

	if (!in_jail(current))
		return 0;

	if (sock->sk->sk_family == AF_UNIX)
		return gtsec_socket_unix_bind(sock, address, addrlen);

	if (address->sa_family != AF_INET)
		return 0;

	tsec = get_security(task,current,security);

	if (!tsec->loc_sectype)
		return 0;

	//if (!tsec->got_network)
	if (!tsec->bsdj_ip)
		/* If we want to be strict, we could just
		 * deny net access when lacking a pseudo ip.
		 * For now we just allow it. */
		return 0;

	inaddr = (struct sockaddr_in *)address;
	sin_addr = inaddr->sin_addr.s_addr;
	jailaddr = tsec->bsdj_ip;

	if (sin_addr == jailaddr)
		return 0;

	if (sin_addr == loopbackaddr || !sin_addr) {
		print_debug(DBG, "Got a loopback or 0 address\n");
		sin_addr = jailaddr;
		print_debug(DBG, "Converted to: %u.%u.%u.%u\n", 
			NIPQUAD(sin_addr));
		return 0;
	}

	return -EPERM;
}

static void
gtsec_socket_post_create(struct socket *sock, int family, int type,
	int protocol)
{
	struct inet_opt *inet;
	struct gtsec_task_sec *tsec;

	if (!in_jail(current))
		return;
	tsec = get_security(task,current,security);
	if (!tsec->loc_sectype)
		return;
	//if (!tsec->got_network)
	if (!tsec->bsdj_ip)
		return;

	if (sock->sk->sk_family != AF_INET)
		return;

	inet = inet_sk(sock->sk);
	inet->saddr = tsec->bsdj_ip;

	return;
}

static int
gtsec_socket_listen(struct socket *sock, int backlog)
{
	struct inet_opt *inet;
	struct gtsec_task_sec *tsec;

	if (!in_jail(current))
		return 0;

	tsec = get_security(task,current,security);
	if (!tsec->loc_sectype)
		return 0;
	//if (!tsec->got_network)
	if (!tsec->bsdj_ip)
		return 0;

	if (sock->sk->sk_family != AF_INET)
		return 0;

	inet = inet_sk(sock->sk);

	if (inet->saddr == tsec->bsdj_ip)
		return 0;

	return -EPERM;
}

#endif	/* CONFIG_SECURITY_NETWORK */

/* for sectype bsdjail mount/umount is not an option */
static int
gtsec_mount(char * dev_name, struct nameidata *nd, char * type,
                         unsigned long flags, void * data)
{
	struct gtsec_task_sec *tsec;

	if (!in_jail(current))
		return 0;

	tsec = get_security(task,current,security);
	if (!tsec->loc_sectype)
		return 0;

	return -EPERM;
}
static int
gtsec_umount(struct vfsmount *mnt, int flags)
{
	struct gtsec_task_sec *tsec;

	if (!in_jail(current))
		return 0;

	tsec = get_security(task,current,security);
	if (!tsec->loc_sectype)
		return 0;

	return -EPERM;
}

/* 
 * process in jail depending on sectype may not:
 *   vnetonly/allbutvnet:
 *     change network config
 *     load/unload modules
 *   bsdjail:
 *     use nice
 *     change network config
 *     load/unload modules
 */
static int
gtsec_capable (struct task_struct *tsk, int cap)
{
	struct gtsec_task_sec *tsec;

	/* exclude for processes in loc_sectype jail: */
	tsec = get_security(task,current,security);
	if (tsec && tsec->in_use && tsec->loc_sectype) {
		if (cap == CAP_NET_ADMIN)
			return -EPERM;
		if (cap == CAP_SYS_MODULE)
			return -EPERM;
		if (cap == CAP_SYS_NICE)
			return -EPERM;
		if (cap == CAP_SYS_RAWIO) {
			return -EPERM;
		}
	}

	/* enforce traditional superuser security */
	if (cap_is_fs_cap (cap) ? tsk->fsuid == 0 : tsk->euid == 0)
		return 0;
	return -EPERM;
}

/*
static int
gtsec_inode_create (struct inode *dir, struct dentry *dentry, int mode)
{
	//printk (KERN_INFO "%s\n", dentry->d_name);

	return 0;
}
*/

static void
gtsec_inode_post_mkdir (struct inode *dir, struct dentry *dentry, int mode)
{
	
	// TODO : watch only within softjail!
	// TODO : race condition: creation and change of i_home ?
	// TODO : if /tmp etc is deleted reset i_tmp to 0 (or whatever)


	/* Check if the parent of the directory create is "/".
	 * TODO: What if rootfs is not ext2/ext3 ?
	 */
	#define EXT3_ROOT_INO                2      /* yanked from ext3_fs.h */
	if (dir->i_ino == EXT3_ROOT_INO) {

		/* Parent directory is "/". */

		if (strcmp(dentry->d_iname,"home") == 0) i_home = dentry->d_inode->i_ino;
		else if (strcmp(dentry->d_iname,"root") == 0) i_root = dentry->d_inode->i_ino;
		else if (strcmp(dentry->d_iname,"tmp") == 0) i_tmp = dentry->d_inode->i_ino;
		else if (strcmp(dentry->d_iname,"dev") == 0) i_dev = dentry->d_inode->i_ino;
		else if (strcmp(dentry->d_iname,"var") == 0) i_var = dentry->d_inode->i_ino;
	}
//	else if (dentry->d_parent->d_parent->d_inode->i_ino == EXT3_ROOT_INO) {
//
//		/* The parent's parent is "/". */
//
//		if ((strcmp(dentry->d_parent->d_iname,"var")) == 0 &&
//				(strcmp(dentry->d_iname,"tmp") == 0)) {
//			i_var = dentry->d_parent->d_inode->i_ino;
//		}
//		
//	}
}

static void
gtsec_inode_post_rename (struct inode *old_dir,
						 struct dentry *old_dentry,
						 struct inode *new_dir,
						 struct dentry *new_dentry) {
	int retval = -EFAULT;
	struct nameidata nd;

	// TODO: If /tmp etc is moved, reset i_tmp to 0.

	if (new_dir->i_ino != EXT3_ROOT_INO) return;

	/* TODO : new_dentry->d_inode is NULL. Why ?
	 * Workaround: redo path_lookup
	 */

	if (strcmp(new_dentry->d_iname,"home") == 0) {
		retval = path_lookup("/home", LOOKUP_FOLLOW | LOOKUP_DIRECTORY, &nd);
		if (retval) return;
		i_home=nd.dentry->d_inode->i_ino;
	}
	else if (strcmp(new_dentry->d_iname,"root") == 0) {
		retval = path_lookup("/root", LOOKUP_FOLLOW | LOOKUP_DIRECTORY, &nd);
		if (retval) return;
		i_root=nd.dentry->d_inode->i_ino;
	}
	else if (strcmp(new_dentry->d_iname,"tmp") == 0) {
		retval = path_lookup("/tmp", LOOKUP_FOLLOW | LOOKUP_DIRECTORY, &nd);
		if (retval) return;
		i_tmp=nd.dentry->d_inode->i_ino;
	}
	else if (strcmp(new_dentry->d_iname,"var") == 0) {
		retval = path_lookup("/var", LOOKUP_FOLLOW | LOOKUP_DIRECTORY, &nd);
		if (retval) return;
		i_var=nd.dentry->d_inode->i_ino;
	}
	else if (strcmp(new_dentry->d_iname,"dev") == 0) {
		retval = path_lookup("/dev", LOOKUP_FOLLOW | LOOKUP_DIRECTORY, &nd);
		if (retval) return;
		i_dev=nd.dentry->d_inode->i_ino;
	}

}


/*
 * This hook replaces inode_post_lookup.
 * It gets called from d_instantiate when an inode is attached to a dentry.
 * Unfortunately, we cannot change which inode gets attached to the dentry here.
 *
 * Gets called when a file/directory is not yet cached,
 * i.e. when a not yet accessed file is accessed
 *      a new file/directory is created.
 * But not when a file/directory is renamed or removed.
 *
 */
/*
static void
gtsec_d_instantiate (struct dentry *dentry, struct inode *inode)
{
	if (!inode) return;	// May be NULL.

	// TODO : Try to use this function instead of the different *_post_* functions.

	printk (KERN_INFO "GTSEC: qqq d_instantiate: %s to %ld\n", dentry->d_iname, inode->i_ino);
}
*/


static int
gtsec_task_alloc_security(struct task_struct *tsk)
{
	struct gtsec_task_sec *tsec;

	if (!in_jail(current))
		return 0;

	/* in jail - child belongs in the same jail */
	tsec = get_security(task,current,security);
	set_security(task,tsk,security,tsec);
	atomic_inc(&tsec->refcount);

	tsec->cur_nrtask++;
	if (tsec->maxtimeslice) {
		tsk->rlim[RLIMIT_CPU].rlim_max = tsec->maxtimeslice;
		tsk->rlim[RLIMIT_CPU].rlim_cur = tsec->maxtimeslice;
	}
	if (tsec->max_data) {
		tsk->rlim[RLIMIT_CPU].rlim_max = tsec->max_data;
		tsk->rlim[RLIMIT_CPU].rlim_cur = tsec->max_data;
	}
	if (tsec->max_memlock) {
		tsk->rlim[RLIMIT_CPU].rlim_max = tsec->max_memlock;
		tsk->rlim[RLIMIT_CPU].rlim_cur = tsec->max_memlock;
	}
	if (tsec->nice)
		set_user_nice(current, tsec->nice);

	return 0;
}

/*
static int
gtsec_bprm_alloc_security(struct linux_binprm *bprm)
{
	struct gtsec_task_sec *tsec;
	//int ret;

	tsec = get_security(task,current,security);
	if (!tsec)
		return 0;

	if (tsec->in_use)
		return 0;

//	if (tsec->net_sectype || tsec->loc_sectype) {
//		printk (KERN_INFO "GTSEC: debug1\n");
//		ret = create_jail(current);
//		if (ret) {
//			// if we failed, nix out the root/ip requests
//			free_task_security(current);
//			return ret;
//		}
//	}
	return 0;
}
*/

static void
gtsec_task_free_security(struct task_struct *tsk)
{
	struct gtsec_task_sec *tsec;

	tsec = get_security(task,tsk,security);
	if (!tsec)
		return;
	if (!tsec->in_use) {
		/* 
		 * someone did 'echo -n x > /proc/<pid>/attr/exec' but
		 * then forked before execing.  Nuke the old info.
		 */
		free_task_security(tsk);
		return;
	}

	tsec->cur_nrtask--;
	/* If this was the last process in the jail, delete the jail */
	if (atomic_dec_and_test(&tsec->refcount)) {
		disable_jail(tsec);
		free_task_security(tsk);
	}
}

/* process in jail may not create devices */
/* TODO: We may want to allow /dev/log, at least... */
static int
gtsec_inode_mknod(struct inode *dir, struct dentry *dentry, int mode, dev_t dev)
{

	if (!in_jail(current))
		return 0;

	if (S_ISFIFO(mode))
		return 0;

	/*
	 * Any harm here ?
	 * Needed for e.g. gconfd-2.
	 */
	if (S_ISSOCK(mode)) {
		print_debug(DBG, "Socket created: %s\n", dentry->d_iname);
		return 0;
	}

	print_debug(DBG, "gtsec: mknod denied for injail task \"%s\".\n", current->comm);
	return -EPERM;
}

/* yanked from fs/proc/base.c */
static unsigned name_to_int(struct dentry *dentry)
{
	const char *name = dentry->d_name.name;
	int len = dentry->d_name.len;
	unsigned n = 0;

	if (len > 1 && *name == '0')
		goto out;
	while (len-- > 0) {
		unsigned c = *name++ - '0';
		if (c > 9)
			goto out;
		if (n >= (~0U-9)/10)
			goto out;
		n *= 10;
		n += c;
	}
	return n;
out:
	return ~0U;
}

/*
 * gtsec_proc_inode_permission:
 *   called only when current is in a jail, and is trying to reach
 *   /proc/<pid>.  We check whether <pid> is in the same jail as
 *   current.  If not, permission is denied.
 */
static int
gtsec_proc_inode_permission(struct inode *inode, int mask,
				    struct nameidata *nd)
{
	struct gtsec_task_sec *tsec = jail_of(current);
	unsigned pid;
	int err = 0;
	struct task_struct *tsk;

	pid = name_to_int(nd->dentry);
	if (pid == ~0U) {
		struct qstr *dname = &nd->dentry->d_name;
		if (strcmp(dname->name, "net")==0 || strcmp(dname->name, "sys")==0 ||
			strcmp(dname->name, "ide")==0)
			return -EPERM;
		return 0;
	}

	read_lock(&tasklist_lock);
	tsk = find_task_by_pid(pid);
	if (tsk && jail_of(tsk) != tsec)
		err = -ENOENT;
	read_unlock(&tasklist_lock);

	return err;
}

/*
 * Here is our attempt to prevent chroot escapes.
 */
static int
is_jailroot_parent(struct dentry *candidate, struct dentry *root,
	struct vfsmount *rootmnt)
{
	if (candidate == root)
		return 0;

	/* simple case:  fs->root/.. == candidate */
	if (root->d_parent == candidate)
		return 1;

	/*
	 * now more complicated:  if fs->root is a mounted directory,
	 * then chdir(..) out of fs->root, at follow_dotdot, will follow
	 * the fs->root mount point. So we must check the parent dir of
	 * the fs->root mount point.
	 */
	if (rootmnt->mnt_root == root && rootmnt->mnt_mountpoint!=root) {
		root = rootmnt->mnt_mountpoint;
		rootmnt = rootmnt->mnt_parent;
		return is_jailroot_parent(candidate, root, rootmnt);
	}

	return 0;
}

static inline short
may_not_even_read(unsigned long ino)
{
	if (ino == i_home) return 1;
	if (ino == i_root) return 1;
	return 0;
}
static int
gtsec_inode_permission(struct inode *inode, int mask,
				    struct nameidata *nd)
{
	struct gtsec_task_sec *tsec;
	struct list_head *tmp;
	struct dentry *d_common;
	struct dentry *dentry;

	if (!in_jail(current))
		return 0;


	tsec = get_security(task,current,security);

	/*
	 * VNETONLY and ALLBUTVNET are not limited here.
	 */
	if (!tsec->loc_sectype)
		return 0;


	/*
	 * In SoftJail Mode
	 *   access to /home, /root is denied and
	 *   write access is limited to /tmp, /var, /dev.
	 */
	if (tsec->loc_sectype == SOFTJAIL) {
		/*
		 * COPY FROM BSDJAIL part. see there.
		 */
		if (nd && nd->dentry &&
			strcmp(nd->dentry->d_sb->s_type->name, "proc")==0) {
			return gtsec_proc_inode_permission(inode, mask, nd);
		}

		#define wants_write ((mask&MAY_WRITE)||(mask&MAY_APPEND))
		#define is_devpts   (strcmp(nd->mnt->mnt_devname,"devpts")==0)
		  // TODO : besser mit procfstype = get_fs_type("devpts"); ??????

		/* Simple things first. */
			/* TODO: Check the mountpoint !
			 * nd->mnt->mnt_mountpoint->d_iname might help!
			 */
		if (wants_write && nd && is_devpts) return 0;

		/*
		 * We need to check all parent inodes of this file/directory.
		 * Additionally, this inode may have several hardlinks. So we need to
		 * check all of them and only grant permission
		 *  - to write, if *all* hardlinks * are pointing to /tmp/... etc.
		 *  - to read/write, if *all* hardlinks * are pointing to /home/... etc.
		 */

		// TODO : use 'list_for_each' or 'list_for_each_safe'

		/* Try to find the first parent inode all hardlinks have in common. */
		tmp=inode->i_dentry.next;
		dentry = list_entry(tmp, struct dentry, d_alias);
		if (S_ISDIR(inode->i_mode)) d_common=dentry;
		else {
			d_common = dentry->d_parent;
			if (may_not_even_read(d_common->d_inode->i_ino)) return -EPERM;

			/* Compare with other hardlinks, if any.
			 * Note, that only files may have hardlinks. So we can skip to the parent
			 * inode, immediately. */
			for (tmp=tmp->next; tmp != &inode->i_dentry; tmp=tmp->next) {
				dentry = list_entry(tmp, struct dentry, d_alias);

				for (dentry=dentry->d_parent; dentry != dentry->d_parent; dentry=dentry->d_parent) {

					if (dentry->d_inode->i_ino == d_common->d_inode->i_ino) break;
					if (may_not_even_read(dentry->d_inode->i_ino)) return -EPERM;
				}
				if (dentry->d_inode->i_ino != d_common->d_inode->i_ino) {
					if (d_common == d_common->d_parent) break;

					/* New d_common and recheck. */
					d_common=d_common->d_parent;
					if (may_not_even_read(d_common->d_inode->i_ino)) return -EPERM;
					tmp=inode->i_dentry.next;
				}
				/* else: We've found a common inode, continue and check the other hardlinks. */
			}
		}

		/* What we've got now:
		 *
		 * d_common is a dentry all hardlinks of this file/directory have in common.
		 *
		 * e.g. an inode with two hardlinks:
		 *   /tmp/subdir/subdir2/file
		 *   /tmp/subdir/otherdir/foobar
		 *   d_common is now dentry for /tmp/subdir/
		 */


		/* All that is left to check is if d_common itself or one of the
		 * parents of d_common actually is a directory we treat specially. */
		for (dentry=d_common; dentry != dentry->d_parent; dentry=dentry->d_parent) {

			/* Check for directories we do not even want to give read access.
			 * At this point common directories are checked, as opposed to above.
			 */
			if (may_not_even_read(dentry->d_inode->i_ino)) return -EPERM;

			/* No special treatment for these, i.e. read/write is okay. */
			if (dentry->d_inode->i_ino == i_tmp && wants_write) return 0;
			if (dentry->d_inode->i_ino == i_var && wants_write) return 0;
			if (dentry->d_inode->i_ino == i_dev && wants_write) return 0;

			/* Currently even write to /proc is denied, so one cannot enter
			 * lantype security mode after activating softjail mode. */
		}
		/* Deny all other write requests. */
		if (wants_write) return -EPERM;

		/* fall through
		 * and grant permission */

	} /* end SOFTJAIL */



	/*
	 * Make some security checks for BSDJAIL !!?
	 */
	if (tsec->loc_sectype == BSDJAIL) {
		if (!nd)
			return 0;

		/*
		 * If trying to get under /proc, we may deny permission:
		 *
		 * Note - we'll want to use sb->s_security to cache whether
		 *  it is the proc fs.  Except that's all the more conflicts
		 *  with selinux security fields.
		 */
		if (nd->dentry &&
			strcmp(nd->dentry->d_sb->s_type->name, "proc")==0) {
			return gtsec_proc_inode_permission(inode, mask, nd);

		}

		/* this is only for 'cd ..' */
		if (!(mask&MAY_EXEC))
			return 0;
		if (!inode || !S_ISDIR(inode->i_mode))
			return 0;
		if (!nd)
			return 0;

		if (is_jailroot_parent(nd->dentry, tsec->dentry, tsec->mnt)) {
			/* you may not chdir(..) out of fs->root */
			print_debug(WARN,"Attempt to chdir(..) out of jail!\n"
					"(%s is a subdir of %s)\n",
					tsec->dentry->d_name.name,
					nd->dentry->d_name.name);
			return -EPERM;
		}

		/* fall through
		 * and grant permission */

	} /* end BSDJAIL */

	/* by default grant permission. */
	return 0;
}

/* process in jail may not send signal to process not in the same jail */
static int
gtsec_task_kill(struct task_struct *p, struct siginfo *info, int sig)
{
	struct gtsec_task_sec *tsec;

	if (!in_jail(current))
		return 0;

	if (jail_of(current) == jail_of(p))
		return 0;

	if (sig==SIGCHLD)
		return 0;

	/*
	 * !?
	 * security.h: "If @info is 1 or SI_FROMKERNEL(info) is true, then the
	 * signal should be viewed as coming from the kernel and should typically
	 * be permitted."
	 *
	 * "@info is 1" meaning ...?
	 * Want to hang your kernel? Then do: if (info && SI_FROMKERNEL(info))
	 */

	/*
	 * TODO: IMPORTANT:
	 * security.h: "SIGIO signals are handled separately by the send_sigiotask hook"
	 * So we are to allow them here ?
	 *
	 * Anyway, X server needs it as wakeup-handler for mouse events.
	 */
	if (sig==SIGIO)
		return 0;

	/* TODO: IMPORTANT:
	 * Dirty workaround: xeyes sends SIGALRM to X server.
	 */
	if (sig==SIGALRM) {
		print_debug(DBG, "gtsec: fix me: sig %d.\n", sig);
		return 0;
	}

	tsec = get_security(task,current,security);
	if (!tsec->loc_sectype)
		return 0;

	print_debug(DBG, "gtsec: sig no. %d denied for injail task \"%s\".\n", sig, current->comm);
	return -EPERM;
}

/*
 * jail_security_task_create:
 *
 * If the current process is in a jail, and that jail is about to exceed a
 * maximum number of processes, then refuse to fork.  If the maximum number
 * of jails is listed as 0, then there is no limit for this jail, and we allow
 * all forks.
 */
static inline int
gtsec_security_task_create (unsigned long clone_flags)
{
	struct gtsec_task_sec *tsec;

	if (!in_jail(current))
		return 0;

	tsec = jail_of(current);

	if (!tsec->loc_sectype)
		return 0;

	if (tsec->max_nrtask && tsec->cur_nrtask >= tsec->max_nrtask)
		return -EPERM;

	return 0;
}

/*
static int
gtsec_inode_readlink(struct dentry *dentry)
{
	return 0;
}
*/

/*
 * LSM hooks to limit jailed process' abilities to muck with resource
 * limits
 */
static int gtsec_task_setrlimit (unsigned int resource, struct rlimit *new_rlim)
{
	struct gtsec_task_sec *tsec;

	if (!in_jail(current))
		return 0;

	/* If no limit set, okay. */
	tsec = jail_of(current);
	if ((tsec->maxtimeslice == 0) && (tsec->max_data == 0) && (tsec->max_memlock == 0))
		return 0;

	return -EPERM;
}

static int gtsec_task_setscheduler (struct task_struct *p, int policy,
				    struct sched_param *lp)
{
	if (!in_jail(current))
		return 0;

	return -EPERM;
}

/*
 * LSM hooks to limit IPC access.
 */

static inline int
basic_ipc_security_check(struct kern_ipc_perm *p, struct task_struct *target)
{
	if (!in_jail(target))
		return 0;

	/*
	 * I think IPC_PRIVATE is harmless ?!:
	 * If a client wants to use a shared memory created with IPC_PRIVATE, it
	 * must be a child process of the server.
	 *
	 * Needed by e.g. openuniverse.
	 *
	 * TODO: shm_alloc_security hook gets not called when needed, so
	 *       p->security is not initialised!
	 *       So we cannot tell, if two tasks are within the same jail.
	 *       This here is a workaround for some programs.
	 */
	if (p->key == IPC_PRIVATE)
		return 0;

	/* This does not work: see above */
	if (p->security != jail_of(target)) {
		print_debug(DBG, "IPC denied.\n");
		return -EPERM;
	}

	return 0;
}

static int
gtsec_ipc_permission(struct kern_ipc_perm *ipcp, short flag)
{
	return basic_ipc_security_check(ipcp, current);
}

static int
gtsec_shm_alloc_security (struct shmid_kernel *shp)
{
	shp->shm_perm.security = jail_of(current);
	return 0;
}

static void
gtsec_shm_free_security (struct shmid_kernel *shp)
{
	shp->shm_perm.security = NULL;
}

static int
gtsec_shm_associate (struct shmid_kernel *shp, int shmflg)
{
	return basic_ipc_security_check(&shp->shm_perm, current);
}

static int
gtsec_shm_shmctl(struct shmid_kernel *shp, int cmd)
{
	if (cmd == IPC_INFO || cmd == SHM_INFO)
		return 0;

	return basic_ipc_security_check(&shp->shm_perm, current);
}

static int
gtsec_shm_shmat(struct shmid_kernel *shp, char *shmaddr, int shmflg)
{
	return basic_ipc_security_check(&shp->shm_perm, current);
}

static int
gtsec_msg_queue_alloc(struct msg_queue *msq)
{
	msq->q_perm.security = jail_of(current);
	return 0;
}

static void
gtsec_msg_queue_free(struct msg_queue *msq)
{
	msq->q_perm.security = NULL;
}

static int
gtsec_msg_queue_associate(struct msg_queue *msq, int flag)
{
	return basic_ipc_security_check(&msq->q_perm, current);
}

static int
gtsec_msg_queue_msgctl(struct msg_queue *msq, int cmd)
{
	if (cmd == IPC_INFO || cmd == MSG_INFO)
		return 0;

	return basic_ipc_security_check(&msq->q_perm, current);
}

static int
gtsec_msg_queue_msgsnd(struct msg_queue *msq, struct msg_msg *msg, int msqflg)
{
	return basic_ipc_security_check(&msq->q_perm, current);
}

static int
gtsec_msg_queue_msgrcv(struct msg_queue *msq, struct msg_msg *msg,
		struct task_struct *target, long type, int mode)

{
	return basic_ipc_security_check(&msq->q_perm, target);
}

static int
gtsec_sem_alloc_security(struct sem_array *sma)
{
	sma->sem_perm.security = jail_of(current);
	return 0;
}

static void
gtsec_sem_free_security(struct sem_array *sma)
{
	sma->sem_perm.security = NULL;
}

static int
gtsec_sem_associate(struct sem_array *sma, int semflg)
{
	return basic_ipc_security_check(&sma->sem_perm, current);
}

static int
gtsec_sem_semctl(struct sem_array *sma, int cmd)
{
	if (cmd == IPC_INFO || cmd == SEM_INFO)
		return 0;

	return basic_ipc_security_check(&sma->sem_perm, current);
}

static int
gtsec_sem_semop(struct sem_array *sma, struct sembuf *sops, unsigned nsops,
	int alter)
{
	return basic_ipc_security_check(&sma->sem_perm, current);
}

/*
 * The next three (socket) hooks prevent a process in a jail from sending
 * data to a abstract unix domain socket which was bound outside the jail.
 */
static int
gtsec_socket_unix_bind(struct socket *sock, struct sockaddr *address,
	int addrlen)
{
	//struct sockaddr_un *sunaddr;
	struct gtsec_task_sec *tsec;

	if (sock->sk->sk_family != AF_UNIX)
		return 0;

	/*
	 * Serge Hallyn has this in his bsdjail LSM.
	 * Why? A socket may certainly have a name. But what has this to do with
	 * being inside a jail or not ?

	sunaddr = (struct sockaddr_un *)address;
	if (sunaddr->sun_path[0] != 0)
		return 0;
	*/

	tsec = jail_of(current);
	set_security(sock,sock->sk,sk_security,tsec);
	return 0;
}

/*
 * Note - we deny sends both from unjailed to jailed, and from jailed
 * to unjailed.  As well as, of course between different jails.
 */
static int
gtsec_socket_unix_may_send(struct socket *sock, struct socket *other)
{
	struct gtsec_task_sec *tsec, *ssec;

	tsec = jail_of(current);  /* jail of sending process */
	ssec = get_security(sock,other->sk,sk_security);  /* jail of receiver */

	if (tsec != ssec) {
		printk (KERN_INFO "Unix datagram connection denied.\n");
		return -EPERM;
	}

	return 0;
}

static int
gtsec_socket_unix_stream_connect(struct socket *sock,
	      struct socket *other, struct sock *newsk)
{
	struct gtsec_task_sec *tsec, *ssec;
	int ret = -EPERM;

	tsec = jail_of(current);  /* jail of sending process */
	ssec = get_security(sock,other->sk,sk_security);  /* jail of receiver */

	if (tsec == ssec) ret = 0;
	else if (tsec->loc_sectype != BSDJAIL) {
		/*
		 * If we simply deny everything here, we would break X socket connections.
		 * So we allow connections to "/tmp/.X11-unix/X0".
		 * This lessens security. But it's called 'soft'-jail, anyway.
		 * A workaround could be to connect to the X server via tcp.
		 */

		//printk (KERN_INFO "gtsec: socket inode: %u.\n", sock_i_ino(other->sk));

		struct unix_sock *u = unix_sk(other->sk);
		unix_state_rlock(other->sk);
		if (u->addr)
		{
			if ((other->file->f_uid == 0) && (other->file->f_gid == 0)) {
				if (strncmp(u->addr->name->sun_path, "/tmp/.X11-unix/X", 16) == 0) /* all X connections */
					ret = 0;
				else if (strcmp(u->addr->name->sun_path, "/dev/log") == 0)	/* e.g.: GConfd */
					ret = 0;
			}
			//else if (strncmp(u->addr->name->sun_path, "/tmp/orbit-", 11) == 0) /* gconf */
			//		ret = 0;
			else
				print_debug(DBG, "Socket connect to %s denied.\n", u->addr->name->sun_path);
		}
		unix_state_runlock(other->sk);
	}

	/*
	if (ret < 0)
		print_debug(DBG, "Socket connect denied.\n");
	*/
	return ret;
}


static struct security_operations gtsec_security_ops = {
	.ptrace						= gtsec_ptrace,
	.capable					= gtsec_capable,

	.task_kill					= gtsec_task_kill,
	.task_alloc_security		= gtsec_task_alloc_security,
	.task_free_security			= gtsec_task_free_security,
	//.bprm_alloc_security		= gtsec_bprm_alloc_security,
	.task_create				= gtsec_security_task_create,

	.task_setrlimit				= gtsec_task_setrlimit,
	.task_setscheduler			= gtsec_task_setscheduler,

	.setprocattr				= gtsec_setprocattr,
	.getprocattr				= gtsec_getprocattr,

	//.inode_create				= gtsec_inode_create,

	.file_set_fowner			= jail_file_set_fowner,
	.file_send_sigiotask		= jail_file_send_sigiotask,

#ifdef CONFIG_SECURITY_NETWORK
	.socket_connect				= gtsec_socket_connect,
	.socket_sendmsg				= gtsec_socket_sendmsg,

	.socket_bind				= gtsec_socket_bind,
	.socket_listen				= gtsec_socket_listen,
	.socket_post_create			= gtsec_socket_post_create,

	.unix_stream_connect		= gtsec_socket_unix_stream_connect,
	.unix_may_send				= gtsec_socket_unix_may_send,
#else
	#error The gtsec module only makes sense when CONFIG_SECURITY_NETWORK is activated.
#endif

	.inode_mknod				= gtsec_inode_mknod,
	.inode_permission			= gtsec_inode_permission,
	.inode_post_mkdir			= gtsec_inode_post_mkdir,
// TODO : sth. like post_rmdir needed to set i_tmp etc. back to undefined.
	.inode_post_rename			= gtsec_inode_post_rename,
	//.inode_readlink				= gtsec_inode_readlink,
	//.d_instantiate				= gtsec_d_instantiate,
	.sb_mount					= gtsec_mount,
	.sb_umount					= gtsec_umount,

	.ipc_permission				= gtsec_ipc_permission,
	.shm_alloc_security			= gtsec_shm_alloc_security,
	.shm_free_security			= gtsec_shm_free_security,
	.shm_associate				= gtsec_shm_associate,
	.shm_shmctl					= gtsec_shm_shmctl,
	.shm_shmat					= gtsec_shm_shmat,

	.msg_queue_alloc_security	= gtsec_msg_queue_alloc,
	.msg_queue_free_security	= gtsec_msg_queue_free,
	.msg_queue_associate		= gtsec_msg_queue_associate,
	.msg_queue_msgctl			= gtsec_msg_queue_msgctl,
	.msg_queue_msgsnd			= gtsec_msg_queue_msgsnd,
	.msg_queue_msgrcv			= gtsec_msg_queue_msgrcv,

	.sem_alloc_security			= gtsec_sem_alloc_security,
	.sem_free_security			= gtsec_sem_free_security,
	.sem_associate				= gtsec_sem_associate,
	.sem_semctl					= gtsec_sem_semctl,
	.sem_semop					= gtsec_sem_semop,
};






///////////////////////////////////////////////////////////////////////////////////////
// BSDJAIL ioctl
///////////////////////////////////////////////////////////////////////////////////////

/*
 * networking ioctl ops:
 * we insert our own wrapper around the dgram and stream ioctl
 * functions, which calls the original ioctl function, then
 * butchers the output so as to show only a jail's own network
 * address.
 */
extern struct proto_ops inet_stream_ops;
extern struct proto_ops inet_dgram_ops;

int (*saved_stream_ioctl)(struct socket *sock, unsigned int cmd,
		unsigned long arg);
int (*saved_dgram_ioctl)(struct socket *sock, unsigned int cmd,
		unsigned long arg);

int jail_stream_ioctl(struct socket *sock, unsigned int cmd, unsigned long arg)
{
	int err = 0;
	struct ifreq ifr;
	struct sockaddr_in *sin;
	struct gtsec_task_sec *tsec = jail_of(current);
	struct ifconf ifc;
	char *lastgood, *cur;
	int oldlen;

	err = saved_stream_ioctl(sock, cmd, arg);

	//if (!tsec || !tsec->in_use || !tsec->got_network)
	if (!tsec || !tsec->in_use || !tsec->bsdj_ip)
		return err;
	if (tsec->loc_sectype != BSDJAIL)
		return err;

	switch (cmd) {
	case SIOCGIFADDR:
		if (copy_from_user(&ifr, (void *)arg, sizeof(struct ifreq)))
			return -EFAULT;
		sin = (struct sockaddr_in *)&ifr.ifr_addr;
		if (sin->sin_family != AF_INET)
			return err;
		if (sin->sin_addr.s_addr != tsec->bsdj_ip) {
			print_debug(WARN, "jail_stream_ioctl DENIED %lu\n",
				(unsigned long)sin->sin_addr.s_addr);
			memset(&ifr, 0, sizeof(struct ifreq));
			copy_to_user((void *)arg, &ifr, sizeof(struct ifreq));
			return -EFAULT;
		}
		break;
	case SIOCGIFCONF:

		print_debug(DBG, "%s called\n", __FUNCTION__);
		if (copy_from_user(&ifc, (void *)arg, sizeof(struct ifconf)))
			return -EFAULT;
		/* first we figure out how much space we really need */
		lastgood = cur = ifc.ifc_buf;
		oldlen = ifc.ifc_len;
		ifc.ifc_len = 0;
		while (cur < ifc.ifc_buf + oldlen) {
			copy_from_user(&ifr, cur, sizeof(struct ifreq));
			sin = (struct sockaddr_in *)&ifr.ifr_addr;
			if (sin->sin_family != AF_INET ||
				sin->sin_addr.s_addr == tsec->bsdj_ip) {
				if (lastgood < cur) {
					copy_to_user(lastgood, &ifr,
						sizeof(struct ifreq));
				}
				ifc.ifc_len += sizeof(struct ifreq);
				lastgood += sizeof(struct ifreq);
				print_debug(DBG, "adding %s\n\n",
					ifr.ifr_name);
			} else {
				print_debug(DBG, "skipping %s\n\n",
					ifr.ifr_name);
			}
			cur += sizeof(struct ifreq);
		}
		memset(&ifr, 0, sizeof(struct ifreq));
		while (lastgood < ifc.ifc_buf + oldlen) {
			copy_to_user(lastgood, &ifr, sizeof(struct ifreq));
			lastgood += sizeof(struct ifreq);
		}
		copy_to_user((void *)arg, &ifc, sizeof(struct ifconf));
	}
	return err;
}

int jail_dgram_ioctl(struct socket *sock, unsigned int cmd, unsigned long arg)
{
	int err = 0;
	struct ifreq ifr;
	struct sockaddr_in *sin;
	struct gtsec_task_sec *tsec = jail_of(current);
	struct ifconf ifc;
	char *lastgood, *cur;
	int oldlen;

	err = saved_dgram_ioctl(sock, cmd, arg);

	//if (!tsec || !tsec->in_use || !tsec->got_network)
	if (!tsec || !tsec->in_use || !tsec->bsdj_ip)
		return err;
	if (tsec->loc_sectype != BSDJAIL)
		return err;

	switch (cmd) {
	case SIOCGIFADDR:
		if (copy_from_user(&ifr, (void *)arg, sizeof(struct ifreq)))
			return -EFAULT;
		sin = (struct sockaddr_in *)&ifr.ifr_addr;
		if (sin->sin_family != AF_INET)
			return err;
		if (sin->sin_addr.s_addr != tsec->bsdj_ip) {
			print_debug(WARN, "jail_dgram_ioctl DENIED %lu\n",
				(unsigned long)sin->sin_addr.s_addr);
			memset(&ifr, 0, sizeof(struct ifreq));
			copy_to_user((void *)arg, &ifr, sizeof(struct ifreq));
			return -EFAULT;
		}
		break;

	case SIOCGIFCONF:
		print_debug(DBG, "%s called\n", __FUNCTION__);
		if (copy_from_user(&ifc, (void *)arg, sizeof(struct ifconf)))
			return -EFAULT;
		/* first we figure out how much space we really need */
		lastgood = cur = ifc.ifc_buf;
		oldlen = ifc.ifc_len;
		ifc.ifc_len = 0;
		while (cur < ifc.ifc_buf + oldlen) {
			copy_from_user(&ifr, cur, sizeof(struct ifreq));
			sin = (struct sockaddr_in *)&ifr.ifr_addr;
			if (sin->sin_family != AF_INET ||
				sin->sin_addr.s_addr == tsec->bsdj_ip) {
				if (lastgood < cur) {
					copy_to_user(lastgood, &ifr,
						sizeof(struct ifreq));
				}
				ifc.ifc_len += sizeof(struct ifreq);
				lastgood += sizeof(struct ifreq);
				print_debug(DBG, "adding %s\n\n",
					ifr.ifr_name);
			} else {
				print_debug(DBG, "skipping %s\n\n",
					ifr.ifr_name);
			}
			cur += sizeof(struct ifreq);
		}
		memset(&ifr, 0, sizeof(struct ifreq));
		while (lastgood < ifc.ifc_buf + oldlen) {
			copy_to_user(lastgood, &ifr, sizeof(struct ifreq));
			lastgood += sizeof(struct ifreq);
		}
		copy_to_user((void *)arg, &ifc, sizeof(struct ifconf));
	}
	return err;
}

void butcher_inet_ops(void)
{
	lock_kernel();
	saved_stream_ioctl = inet_stream_ops.ioctl;
	saved_dgram_ioctl = inet_dgram_ops.ioctl;
	inet_stream_ops.ioctl = jail_stream_ioctl;
	inet_dgram_ops.ioctl = jail_dgram_ioctl;
	unlock_kernel();
}

void unbutcher_inet_ops(void)
{
	lock_kernel();
	inet_stream_ops.ioctl = saved_stream_ioctl;
	inet_dgram_ops.ioctl = saved_dgram_ioctl;
	unlock_kernel();
}

///////////////////////////////////////////////////////////////////////////////////////
// END BSDJAIL ioctl
///////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////
// /proc/gtsec/ hooks
///////////////////////////////////////////////////////////////////////////////////////

/* Needed to show the destination when doing a ls /proc/gtsec */
static
int proc_readlink_homesymlink(struct dentry *dentry, char __user *buffer, int buflen)
{
	int len;
	struct gtsec_task_sec *tsec;
	char *ptr;

	ptr = "/tmp";
	if (in_jail(current)) {
		tsec = get_security(task,current,security);
		if (tsec->loc_sectype == SOFTJAIL)
			ptr = tsec->sj_home;
	}

	len = strlen(ptr);

	if (copy_to_user(buffer, ptr, len))
		return -EFAULT;
	return len;
}
/* Needed to chdir to destination when doing a cd /proc/gtsec/home */
static
int proc_followlink_homesymlink (struct dentry *dentry, struct nameidata *nd)
{
	struct gtsec_task_sec *tsec;

	if (in_jail(current)) {
		tsec = get_security(task,current,security);
		if (tsec->loc_sectype == SOFTJAIL)
			return vfs_follow_link(nd, tsec->sj_home);
	}

	// TODO: maybe return real home dir?
	return vfs_follow_link(nd,"/tmp");
}
static struct inode_operations proc_iops_homesymlink = {
	readlink:			proc_readlink_homesymlink,
	follow_link:		proc_followlink_homesymlink,

};


/* Needed to show the destination when doing a ls /proc/gtsec */
static
int proc_readlink_passwdsymlink(struct dentry *dentry, char __user *buffer, int buflen)
{
	int len;
	struct gtsec_task_sec *tsec;
	char *ptr;

	ptr = ORIGINAL_ETC_PASSWD;
	if (in_jail(current)) {
		tsec = get_security(task,current,security);
		if (tsec->loc_sectype == SOFTJAIL)
			ptr = "passwd.softjail";
	}


	len = strlen(ptr);

	if (copy_to_user(buffer, ptr, len))
		return -EFAULT;
	return len;
}

/* Needed to chdir to destination when doing a cd /proc/gtsec/home */
static
int proc_followlink_passwdsymlink(struct dentry *dentry, struct nameidata *nd)
{
	struct gtsec_task_sec *tsec;

	if (in_jail(current)) {
		tsec = get_security(task,current,security);
		if (tsec->loc_sectype == SOFTJAIL)
			return vfs_follow_link(nd,"passwd.softjail");
	}

	return vfs_follow_link(nd,ORIGINAL_ETC_PASSWD);

}
static struct inode_operations proc_iops_passwdsymlink = {
	readlink:			proc_readlink_passwdsymlink,
	follow_link:		proc_followlink_passwdsymlink,

};


/*
 * Where do we get the passwd-file data from?
 * Best way I can figure is to
 *   cat /etc/passwd.original > /proc/gtsec/passwd.softjail (by root)
 * shortly after loading the module.
 *
 * Downside of course is, that there is no data in /etc/passwd if gtsec is
 * not loaded, so one can't even log in as superuser!
 * Suggestions ?
 * I do not see a reasonable way to intercept the inode lookup of /etc/passwd.
 *
 * Documentation for seq_file see:
 * http://lwn.net/Articles/22355/
 * http://lwn.net/Articles/23707/
 * http://www.diku.dk/hjemmesider/studerende/firefly/emspace-html/node3.html
 */
static ssize_t proc_write_passwdfile(struct file *file,
									//char __user *buf,
									const char *buf,
									size_t size,
									loff_t *ppos)
{
	char *tmp=NULL;

	/* May not be changed within jail! */
	if (in_jail(current))
		return -EPERM;

	// TODO : What about multiple tasks writing to /proc/gtsec/passwd.softjail ?

	if (*ppos == 0) {
		if (passwd_data) {
			kfree(passwd_data);
			passwd_data = NULL;
		}

		// TODO : For maximum memory waste, we don't do any dynamic reallocation here.
		print_debug(DBG, "gtsec: allocating %d bytes\n", max_passwd_len);
		passwd_data = kmalloc(max_passwd_len, GFP_KERNEL);
		if (!passwd_data)
			return -ENOMEM;
	}
	tmp=passwd_data;

	/* Check for buffer overrun */
	if ((*ppos+size) >= max_passwd_len) {
		printk (KERN_INFO "gtsec: Too much data. Try to increase max_passwd_len in lsm source.\n");
		return -EFAULT;
	}
	tmp += *ppos;
  
	write_lock(&place_lock);		// TODO: implementation correct ?
	if (copy_from_user(tmp, buf, size)) {
		write_unlock(&place_lock);
		return -EFAULT;
	}
	tmp[size]='\0';
	write_unlock(&place_lock);

	*ppos += size;
	return size;
}
static int proc_show_passwdfile(struct seq_file *m, void *v)
{
	if (in_jail(current)) {
		struct gtsec_task_sec *tsec;
		tsec = get_security(task,current,security);
		if (tsec->loc_sectype == SOFTJAIL) {
			/*
			 * Print the customized user entry.
			 * TODO: delete the duplicate in passwd_dat
			 * TODO: get correct login shell
			 */
			seq_printf(m, "%s:x:%d:%d:%s,,,:%s:/bin/bash\n%s",
					tsec->sj_username,
					tsec->sj_uid, tsec->sj_gid,
					tsec->sj_username,
					tsec->sj_home,
					passwd_data?passwd_data:"");
		}

		return 0;
	}

	if (!passwd_data) {
		/*
		 * Make sure at least root can log in, even if no data was pushed to
		 * /proc/gtsec/passwd.softjail.
		 */
		seq_printf(m, "root:x:0:0:root:/root:/bin/bash\n");
		return 0;
	}

	seq_printf(m, passwd_data);
	return 0;
}
static
int proc_open_passwdfile (struct inode *inode, struct file *file) {
	return single_open(file, proc_show_passwdfile, NULL);
}
static struct file_operations proc_fops_passwdfile = {
	.owner   = THIS_MODULE,
	.open    = proc_open_passwdfile,
	.read    = seq_read,
	.write   = proc_write_passwdfile,
	.llseek  = seq_lseek,
	.release = single_release,
};

static
void setup_gtsec_proc(void) {

	gtsec_proc_dir = proc_mkdir(MY_NAME, NULL);
	//gtsec_proc_dir = proc_mkdir(MY_NAME, &proc_root);
	if(gtsec_proc_dir == NULL) {
		printk (KERN_INFO "gtsec: Could not create /proc/gtsec\n");
		return;
	}
	gtsec_proc_dir->owner = THIS_MODULE;

	home_symlink = create_proc_entry("home", S_IFLNK|S_IRWXUGO, gtsec_proc_dir);
	if(home_symlink == NULL) {
		printk (KERN_INFO "gtsec: Could not create /proc/gtsec/home\n");
		return;
	}
	home_symlink->proc_iops = &proc_iops_homesymlink;
	home_symlink->owner = THIS_MODULE;

	passwd_symlink = create_proc_entry("passwd", S_IFLNK|S_IRWXUGO, gtsec_proc_dir);
	if(passwd_symlink == NULL) {
		printk (KERN_INFO "gtsec: Could not create /proc/gtsec/passwd\n");
		return;
	}
	passwd_symlink->proc_iops = &proc_iops_passwdsymlink;
	passwd_symlink->owner = THIS_MODULE;

	passwd_file = create_proc_entry("passwd.softjail", 0644, gtsec_proc_dir);
	if(passwd_file == NULL) {
		printk (KERN_INFO "gtsec: Could not create /proc/gtsec/passwd.softjail\n");
		return;
	}
	passwd_file->proc_fops = &proc_fops_passwdfile;
	//passwd_file->read_proc = proc_read_passwd;
	//passwd_file->write_proc = proc_write_passwd;
	passwd_file->owner = THIS_MODULE;

}

static
void unsetup_gtsec_proc(void) {
	remove_proc_entry("passwd", gtsec_proc_dir);

	remove_proc_entry("passwd.softjail", gtsec_proc_dir);
	if (passwd_data) kfree(passwd_data);

	remove_proc_entry("home", gtsec_proc_dir);
	remove_proc_entry(MY_NAME, &proc_root);
}

///////////////////////////////////////////////////////////////////////////////////////////
// END /proc/gtsec hooks
///////////////////////////////////////////////////////////////////////////////////////////











static int __init gtsec_init (void)
{

	setup_gtsec_proc();
	butcher_inet_ops();

	if (register_security (&gtsec_security_ops)) {
		printk (KERN_INFO 
			"Failure registering GTSEC module with the kernel\n");

		if (mod_reg_security (MY_NAME, &gtsec_security_ops)) {
			printk (KERN_INFO "Failure registering GTSEC"
				" module with primary security module.\n");
			return -EINVAL;
		}
		secondary = 1;
	}
	printk (KERN_INFO "GTSEC module initialized.\n");

	return 0;
}

static void __exit gtsec_exit (void)
{
	unsetup_gtsec_proc();
	unbutcher_inet_ops();

	if (secondary) {
		if (mod_unreg_security (MY_NAME, &gtsec_security_ops))
			printk (KERN_INFO "Failure unregistering GTSEC"
				" module with primary module.\n");
	} else { 
		if (unregister_security (&gtsec_security_ops)) {
			printk (KERN_INFO "Failure unregistering GTSEC"
				"module with the kernel\n");
		}
	}

	printk (KERN_INFO "GTSEC module removed\n");
}

security_initcall (gtsec_init);
module_exit (gtsec_exit);

MODULE_DESCRIPTION("GT Security Module");
MODULE_LICENSE("GPL");
